"""multistate joint modeling package.

This package provides tools for multistate joint modeling with PyTorch.
"""

from .model import MultiStateJointModel

__author__ = "Félix Laplante"
__email__ = "felixlaplante0@gmail.com"
__license__ = "MIT"

__all__ = ["MultiStateJointModel"]
